#!/usr/bin/python

#
#  Copyright (C) 2004 Red Hat, Inc.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import gobject
import os
import socket
import shutil
import re
import errno
import rpm
import lvm
import TransferMethod
import LDAPUtils

from config import *
from statelessGenPXEConfig import generate_pxelinux_config

def mkdir (dir):
    try:
        os.makedirs (dir)
    except OSError, err:
        if err.errno != errno.EEXIST:
            raise err

def hostname ():
    try:
        return socket.gethostbyname(socket.gethostname())
    except:
        return socket.gethostname()

def get_readonly_service ():
    cfg = StatelessConfig ()
    return LDAPUtils.StatelessService (LDAPUtils.ReadOnlyDirectory (cfg))

def list_protosystems ():
    protosystems = []
    srv = get_readonly_service ()
    config_list = srv.get_configuration_list ()
    for entry in config_list:
        protosystems.append (entry.get_attr_value ("statelessConfigurationName"))
    return protosystems

def list_snapshots (protosystem):
    snapshots = []
    srv = get_readonly_service ()
    snapshot_list = srv.get_snapshot_list (protosystem)
    for entry in snapshot_list:
        snapshots.append (entry.get_attr_value ("statelessSnapshotName"))
    return snapshots

def get_next_id (protosystem):
    ret = 0
    r = re.compile ('%s-([\d]+)' % protosystem)
    for snapshot in list_snapshots (protosystem):
        m = r.match (snapshot)
        if m:
            tmpid = int (m.groups()[0])
            if tmpid > ret:
                ret = tmpid
    return ret + 1

# We can only take LVM snapshots of a src_path if root underneath it is a mountpoint of an LVM Logical Volume
def can_support_lvm_snapshot (src_path):
    if lvm.is_lvm_logical_volume(src_path):
        if lvm.kernel_supports_snapshots():
            # print "Can use LVM snapshot technique: the protosystem is a mountpoint of an LVM logical volume"
            return True
    else:
        #print "Cannot use LVM snapshot technique: the protosystem is not a mountpoint of an LVM logical volume"
        return False

def mount(device_name, mount_point):
    command = "%s %s %s" % ('/bin/mount', device_name, mount_point)
    result = os.system (command)
    if result != 0:
        raise SystemError

class LVMSnapshotMethod (TransferMethod.TransferMethod):
    # this can only work if src is a mounted LVM volume
    # the directory dest should already exist...
    def __init__ (self, src_lv, dest_dir, snapshot_lv_name):
        self.src_lv = src_lv
        self.dest_dir = dest_dir
        self.snapshot_lv_name = snapshot_lv_name
        TransferMethod.TransferMethod.__init__ (self)

    def do_start_transfer (self):
        size_string = "-L1G"  # FIXME
        snapshot_volume_device = "/dev/%s/%s"%(self.src_lv.vg_name, self.snapshot_lv_name)
            
        lvcreate_command = "lvcreate %s -s -n %s %s"%(size_string, self.snapshot_lv_name, self.src_lv.lv_name)
        print lvcreate_command
        
        # FIXME: should have better error detection:
        if 0!=os.system(lvcreate_command):
            return False

        self.update (50.0)

        # Mount the snapshot volume at the dst directory...
        # FIXME: should we add it to fstab?
        mount(snapshot_volume_device, self.dest_dir)

        self.update (100.0)
        
        self.transfer_finished (True)

class Snapshooter (gobject.GObject):
    __gsignals__ = {
        'snapshot-finished' : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
        }
    
    # List of rpms that we must have installed.
    required_rpms = ['readonly-root',
                     'dhclient',
                     'rsync',
                     'stateless-client'
                     ]

    def __init__ (self, protosystem, serial):
        gobject.GObject.__init__ (self)
        self.protosystem = protosystem
        self.serial = serial
        self.snapshot_name = self.protosystem + '-' + str(self.serial)

        self.cfg = StatelessConfig ()
        self.server = self.cfg ["SERVER_NAME"]
        if self.server == None:
            self.server = hostname ()

        self.directory = LDAPUtils.ReadWriteDirectory (self.cfg)
        self.stateless = LDAPUtils.StatelessService (self.directory)

        # Set up the destination directory:
        snapshots_path = os.path.normpath ('/'.join ([SNAPSHOTSBASE, self.protosystem]))
        mkdir (snapshots_path)
        
        self.snapshot_path = '/'.join ([snapshots_path,
                                        self.protosystem + '-' + str (self.serial)])
        mkdir (self.snapshot_path)

        protosystem_path = os.path.normpath ('/'.join ([PROTOSYSTEMSBASE, self.protosystem]))

        # Use appropriate sync method:
        if can_support_lvm_snapshot (protosystem_path):
            #print "Using LVM to take snapshot"
            lv = lvm.get_lv_mounted_at (protosystem_path)
            snapshot_lv_name = self.protosystem +'_' + str (self.serial)
            self.transfer = LVMSnapshotMethod (lv, self.snapshot_path + '/', snapshot_lv_name)
        else:
            #print "Using rsync to take snapshot"
            self.transfer = TransferMethod.RsyncMethod (protosystem_path + '/', self.snapshot_path + '/')
        self.transfer.connect ('transfer-finished', self.transfer_finished)

    def get_transfer (self):
        return self.transfer

    def transfer_finished (self, transfer, success):
        self.prepare_destination()

    def prepare_readonly_root (self):
        os.system ("chroot %s /usr/sbin/readonly-prepare /" % self.snapshot_path)

    def clear_rpm_db_files (self):
        """Removes rpm database files in the destination if they exist"""
        def rmpath (p):
            try: os.remove (p)
            except OSError: pass
        rmpath (os.path.join (self.snapshot_path, 'var/lib/rpm/__db.001'))
        rmpath (os.path.join (self.snapshot_path, 'var/lib/rpm/__db.002'))
        rmpath (os.path.join (self.snapshot_path, 'var/lib/rpm/__db.003'))

    def check_installed_rpms (self):
        """Sanity check the target system for required RPMs"""
        # Always clear out the db files before and after doing any operations
        self.clear_rpm_db_files ()
        ts = rpm.TransactionSet (self.snapshot_path)
        for package in self.required_rpms:
            mi = ts.dbMatch ('name', package)
            if mi.count () == 0:
                print package + ' is missing...'
        self.clear_rpm_db_files ()

    def get_latest_kernel (self):
        kernel_version = None
        self.clear_rpm_db_files ()
        ts = rpm.TransactionSet (self.snapshot_path)
        for mi in ts.dbMatch ('name', 'kernel'):
            kernel_version = "%s-%s" % (mi['version'], mi['release'])
        self.clear_rpm_db_files ()
        return kernel_version

    def prepare_diskless_initrd (self):
        latest_kernel = self.get_latest_kernel ()
        os.system ("chroot %s /usr/sbin/diskless-mkinitrd -f /boot/diskless-initrd-%s.img %s" %
                   (self.snapshot_path, latest_kernel, latest_kernel))
    
    def write_snapshot_info (self):
        def write_snapshot_info_to_file (path):
            f = file (path, 'w')
            f.write ("STATELESS_CONFIGURATION=%s\n" % self.protosystem)
            f.write ("STATELESS_SNAPSHOT=%s\n" % self.snapshot_name)
            f.close ()
            os.chmod (path, 0644)

        path = os.path.normpath (self.snapshot_path + '/etc/sysconfig/stateless-snapshot')
        write_snapshot_info_to_file (path)
        path = os.path.normpath (self.snapshot_path + '/boot/stateless-snapshot')
        write_snapshot_info_to_file (path)

    def write_stateless_config (self):
        cfg = StatelessConfig (self.snapshot_path)
        cfg["LDAP_SERVER"]  = self.cfg["LDAP_SERVER"]
        cfg["LDAP_BASE_DN"] = self.cfg["LDAP_BASE_DN"]
        cfg.write (0077)
        
    def update_latest_version_info (self):
        if self.stateless.get_configuration(self.protosystem) == None:
            self.stateless.add_configuration(self.protosystem, self.snapshot_name)
        else:
            self.stateless.set_default_snapshot(self.protosystem, self.snapshot_name)
        self.stateless.add_snapshot(self.protosystem, self.snapshot_name, [self.server])

    def copy_pxeboot_files (self):
        def copy_from_snapshot (src, dest):
            src_path = os.path.normpath ('/'.join ([self.snapshot_path, 'boot', src]))
            dest_path = os.path.normpath ('/'.join ([TFTPBASE, 'stateless', self.protosystem, self.snapshot_name, dest]))
            shutil.copy (src_path, dest_path)
            
        mkdir (os.path.normpath ('/'.join ([TFTPBASE, 'stateless', self.protosystem, self.snapshot_name])))
        
        latest_kernel = self.get_latest_kernel ()
        copy_from_snapshot ("diskless-initrd-%s.img" % latest_kernel, 'initrd.img')
        copy_from_snapshot ("vmlinuz-%s" % latest_kernel, 'vmlinuz')

    def enable_stateless_update (self):
        os.system ("chroot %s /sbin/chkconfig --level 345 stateless on" % self.snapshot_path)
    
    def prepare_destination (self):
        self.prepare_readonly_root ()
        self.prepare_diskless_initrd ()
        self.check_installed_rpms ()
        self.write_snapshot_info ()
        self.write_stateless_config ()
        self.copy_pxeboot_files ()
        self.enable_stateless_update ()
        self.update_latest_version_info ()
        generate_pxelinux_config ()
        self.emit ('snapshot-finished')

    def start (self):
        self.transfer.start_transfer ()
        
gobject.type_register (Snapshooter)
